#include "CuttingStockApp.h"
#include "cuttingAlgorithm.h"
#include "BlfWorker.h"
#include <QThread>
#include <QFileDialog>
#include <boost/filesystem/path.hpp>

using boost::filesystem::path;

namespace cutting_stock
{
	const char* CuttingStockApp::TITLE = "Cutting Stock 2D";
	const char* CuttingStockApp::PARAMS = "./config/parameters.txt";
	const char* CuttingStockApp::INPUT_PATH = "./input/";

	static const char* statusMessages[] = {
		"Algoritam je inicijaliziran.",
		"Algoritam se izvr\305\241ava. Trenutna generacija: " ,
		"Algoritam se izvr\305\241io.",
		"Ulazna datoteka je uspje\305\241no u\304\215itana.",
		"Izlazna datoteka je uspje\305\241no u\304\215itana.",
		"Do\305\241lo je do pogre\305\241ke pri inicijalizaciji algoritma.",
		"Do\305\241lo je do pogre\305\241ke pri \304\215itanju ulazne datoteke.",
		"Do\305\241lo je do pogre\305\241ke pri \304\215itanju izlazne datoteke."
	};

	enum Status
	{
		INIT,
		EXECUTING,
		SUCCESS,
		SUCCESS_INPUT,
		SUCCESS_OUTPUT,
		ERR_INIT,
		ERR_INPUT,
		ERR_OUTPUT
	};

	CuttingStockApp::CuttingStockApp(QWidget *parent) : QMainWindow(parent)
	{
		ui.setupUi(this);
		init();
	}

	CuttingStockApp::~CuttingStockApp()
	{
		delete colours;
	}

	void CuttingStockApp::init() {
		// open params xml and set text field values
		XMLNode xConfig = XMLNode::openFileHelper(PARAMS, "ECF");

		XMLNode registry = xConfig.getChildNode("Registry");
		setTextFieldFromNode(ui.populationSize, registry, "population.size");
		//setTextFieldFromNode(ui.Iterations, registry, "batch.repeats");
		ui.Iterations->setText(QString::number(iterations));

		XMLNode permutation = xConfig.getChildNode("Genotype").getChildNode("Permutation");
		setTextFieldFromNode(ui.COSA, permutation, "crx.COSA");
		setTextFieldFromNode(ui.DPX, permutation, "crx.DPX");
		setTextFieldFromNode(ui.OBX, permutation, "crx.OBX");
		setTextFieldFromNode(ui.OPX, permutation, "crx.OPX");
		setTextFieldFromNode(ui.OX, permutation, "crx.OX");
		setTextFieldFromNode(ui.OX2, permutation, "crx.OX2");
		setTextFieldFromNode(ui.PBX, permutation, "crx.PBX");
		setTextFieldFromNode(ui.PMX, permutation, "crx.PMX");
		setTextFieldFromNode(ui.SPX, permutation, "crx.SPX");
		setTextFieldFromNode(ui.ULX, permutation, "crx.ULX");
		setTextFieldFromNode(ui.UPMX, permutation, "crx.UPMX");
		setTextFieldFromNode(ui.cyclic, permutation, "crx.cyclic");
		setTextFieldFromNode(ui.cyclic2, permutation, "crx.cyclic2");
		setTextFieldFromNode(ui.m_ins, permutation, "mut.ins");
		setTextFieldFromNode(ui.m_inv, permutation, "mut.inv");
		setTextFieldFromNode(ui.m_toggle, permutation, "mut.toggle");
		dimension = QString(permutation.getChildNodeWithAttribute("Entry", "key", "size").getText()).toInt();

		XMLNode func = registry.getChildNodeWithAttribute("Entry", "key", "tsp.infile");
		inputFile = func.getText();

		// set title and status bar
		setWindowTitleFromFile(inputFile);
		statusMessage = new QLabel;
		ui.statusBar->addWidget(statusMessage);
		statusMessage->setText(statusMessages[INIT]);

		// set scene
		colours = new PolygonView(ui.graphicsView);
		colours->resize(ui.graphicsView->maximumWidth(), ui.graphicsView->maximumHeight());

		// connect signals and slots
		connect(ui.actionLoad, SIGNAL(triggered()), this, SLOT(openFile()));
		connect(ui.loadOutputButton, SIGNAL(clicked()), this, SLOT(openOutput()));		
		connect(ui.prevButton, SIGNAL(clicked()), this, SLOT(prevSolution()));
		connect(ui.nextButton, SIGNAL(clicked()), this, SLOT(nextSolution()));
		connect(ui.executeButton, SIGNAL(clicked()), this, SLOT(startExecute()));

		ui.prevButton->setDisabled(true);
		ui.nextButton->setDisabled(true);

		drawInitialSolution();
	}

	void CuttingStockApp::getSolution(const QVector<GraphicsStructP> &graphicsVector) {
		this->graphicsVector = graphicsVector.toStdVector();

		currentIndex = 0;
		ui.prevButton->setEnabled(false);
		if (this->graphicsVector.size() > 1) {
			ui.nextButton->setEnabled(true);
		}
		else ui.nextButton->setEnabled(false);

		// iscrtaj prvi
		drawSolution();
		setFitness();

		statusMessage->setText(statusMessages[SUCCESS]);
		ui.executeButton->setEnabled(true);

		// vrati broj iteracija u config
		//iterations = ui.Iterations->text().toInt();
		//updateXml();
	}

	void CuttingStockApp::startExecute() {
		ui.executeButton->setEnabled(false);

		//iterations = 1;
		iterations = ui.Iterations->text().toInt();
		updateXml();

		// asinkrono
		auto thread = new QThread;
		const path filePath(inputFile.c_str());
		auto worker = new BlfWorker(iterations, filePath.filename().string());
		worker->moveToThread(thread);

		qRegisterMetaType<QVector<GraphicsStructP>>();
		connect(worker, SIGNAL(error(QString)), this, SLOT(errorString(QString)));
		connect(thread, SIGNAL(started()), worker, SLOT(process()));
		connect(worker, SIGNAL(setSolution(QVector<cutting_stock_graphics::GraphicsStructP>)), 
				this, SLOT(getSolution(QVector<cutting_stock_graphics::GraphicsStructP>)));
		connect(worker, SIGNAL(finished()), thread, SLOT(quit()));
		connect(worker, SIGNAL(finished()), worker, SLOT(deleteLater()));
		connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
		connect(worker, SIGNAL(sendGeneration(QString)), this, SLOT(setStatusText(QString)));

		thread->start();
	}

	void CuttingStockApp::drawInitialSolution() {
		char *arg0, *arg1;
		const int argc = getArgs(arg0, arg1);
		char* argv[] = { arg0, arg1, nullptr };

		const CuttingStockEvalOpP tsp(new CuttingStockEvalOp());
		StateP state(new State);
		state->setEvalOp(tsp);
		const bool initSuccess = state->initialize(argc, argv);

		delete[] arg0;
		delete[] arg1;

		if (!initSuccess) {
			statusMessage->setText(statusMessages[ERR_INIT]);
			return;
		}

		vector<int> elements;
		for (uint i = 0; i < dimension; ++i) {
			elements.push_back(i);
		}

		graphicsVector.clear();
		const auto solution = getBlfSolution(tsp, elements);
		graphicsVector.push_back(solution);

		currentIndex = 0;
		ui.prevButton->setEnabled(false);
		if (this->graphicsVector.size() > 1) {
			ui.nextButton->setEnabled(true);
		}

		drawSolution();
	}

	void CuttingStockApp::drawSolution() {
		const auto graphics = graphicsVector[currentIndex].get();
		colours->elements = graphics->elements;
		colours->width = graphics->width;
		colours->ymax = graphics->ymax;
		colours->canvasHeight = ui.graphicsView->height();
		colours->repaint();
	}

	void CuttingStockApp::setFitness()
	{
		const auto graphics = graphicsVector[currentIndex].get();
		setStatusMessageFromFitness(graphics->fitness);
	}

	void CuttingStockApp::openFile() {
		inputFile = QFileDialog::getOpenFileName(this, tr("Open input file"), tr(INPUT_PATH), tr("Input files (*.tsp *.txt)")).toStdString();

		ifstream fin(inputFile);
		if (fin.is_open()) {
			string line;
			while (getline(fin, line) && line.find("size") == string::npos);
			stringstream ss(line.substr(line.find(":") + 1));
			ss >> dimension;
			fin.close();
		}
		else {
			statusMessage->setText(statusMessages[ERR_INPUT]);
			return;
		}

		// update tsp infile
		updateXml();

		setWindowTitleFromFile(inputFile);
		statusMessage->setText(statusMessages[SUCCESS_INPUT]);

		drawInitialSolution();
	}

	void CuttingStockApp::openOutput() {
		const auto outputFile = QFileDialog::getOpenFileName(this,
			tr("Open output file"), "./output/", tr("Output files (*.out)")).toStdString();

		ifstream fin(outputFile);
		if (!fin.is_open()) {
			statusMessage->setText(statusMessages[ERR_OUTPUT]);
			return;
		}

		char *arg0, *arg1;
		const int argc = getArgs(arg0, arg1);
		char* argv[] = { arg0, arg1, nullptr };

		const CuttingStockEvalOpP tsp(new CuttingStockEvalOp());
		StateP state(new State);
		state->setEvalOp(tsp);
		const bool initSuccess = state->initialize(argc, argv);

		delete[] arg0;
		delete[] arg1;

		if (!initSuccess) {
			statusMessage->setText(statusMessages[ERR_INIT]);
			fin.close();
			return;
		}

		graphicsVector.clear();
		string line;
		while (getline(fin, line)) {
			int x;
			vector<int> elements;
			istringstream iString(line);
			while (iString >> x) {
				elements.push_back(x);
			}
			const auto solution = getBlfSolution(tsp, elements);
			graphicsVector.push_back(solution);
		}
		fin.close();

		currentIndex = 0;
		ui.prevButton->setEnabled(false);
		if (this->graphicsVector.size() > 1) {
			ui.nextButton->setEnabled(true);
		}

		statusMessage->setText(statusMessages[SUCCESS_OUTPUT]);

		// iscrtaj prvi
		drawSolution();
		setFitness();
	}

	void CuttingStockApp::prevSolution() {
		ui.nextButton->setEnabled(true);
		if (--currentIndex == 0) {
			ui.prevButton->setEnabled(false);
		}

		drawSolution();
		setFitness();
	}

	void CuttingStockApp::nextSolution() {
		ui.prevButton->setEnabled(true);
		const int lastIndex = graphicsVector.size() - 1;
		if (++currentIndex == lastIndex) {
			ui.nextButton->setEnabled(false);
		}

		drawSolution();
		setFitness();
	}

	void CuttingStockApp::setWindowTitleFromFile(const string& file)
	{
		const path filePath(file.c_str());
		setWindowTitle((string(TITLE) + " - " + filePath.filename().string()).c_str());
	}

	int CuttingStockApp::getArgs(char*& arg0, char*& arg1)
	{
		arg0 = new char[strlen(TITLE) + 1];
		arg1 = new char[strlen(PARAMS) + 1];

		strcpy(arg0, TITLE);
		strcpy(arg1, PARAMS);

		char* argv[] = { arg0, arg1, nullptr };
		const int argc = sizeof(argv) / sizeof(argv[0]) - 1;

		return argc;
	}

	void CuttingStockApp::setStatusText(const QString& genNo) const {
		statusMessage->setText(statusMessages[EXECUTING] + genNo);
	}

	void CuttingStockApp::setStatusMessageFromFitness(const double& fitness) const
	{
		ui.fitnessLabel->setText("Fitness: " + QString::number(fitness) + "%");
	}

	void CuttingStockApp::parseXml(string fileName) {
		ifstream fin(fileName);
		if (fin.is_open()) {
			string line;
			while (getline(fin, line))
				xmlFile += line += "\n";
			fin.close();
		}
		else throw runtime_error("Cannot open parameters file.");
	}

	void CuttingStockApp::updateXml() const
	{
		XMLNode xConfig = XMLNode::openFileHelper(PARAMS, "ECF");

		XMLNode registry = xConfig.getChildNode("Registry");
		registry.getChildNodeWithAttribute("Entry", "key", "tsp.infile").updateText(inputFile.c_str());

		updateXmlNodeFromTextField(ui.populationSize, registry, "population.size");

		//XMLNode batchRepeats = registry.getChildNodeWithAttribute("Entry", "key", "batch.repeats");
		//batchRepeats.updateText(QString::number(iterations).toStdString().c_str());

		XMLNode permutation = xConfig.getChildNode("Genotype").getChildNode("Permutation");
		XMLNode size = permutation.getChildNodeWithAttribute("Entry", "key", "size");
		size.updateText(QString::number(dimension).toStdString().c_str());
		updateXmlNodeFromTextField(ui.COSA, permutation, "crx.COSA");
		updateXmlNodeFromTextField(ui.DPX, permutation, "crx.DPX");
		updateXmlNodeFromTextField(ui.OBX, permutation, "crx.OBX");
		updateXmlNodeFromTextField(ui.OPX, permutation, "crx.OPX");
		updateXmlNodeFromTextField(ui.OX, permutation, "crx.OX");
		updateXmlNodeFromTextField(ui.OX2, permutation, "crx.OX2");
		updateXmlNodeFromTextField(ui.PBX, permutation, "crx.PBX");
		updateXmlNodeFromTextField(ui.PMX, permutation, "crx.PMX");
		updateXmlNodeFromTextField(ui.SPX, permutation, "crx.SPX");
		updateXmlNodeFromTextField(ui.ULX, permutation, "crx.ULX");
		updateXmlNodeFromTextField(ui.UPMX, permutation, "crx.UPMX");
		updateXmlNodeFromTextField(ui.cyclic, permutation, "crx.cyclic");
		updateXmlNodeFromTextField(ui.cyclic2, permutation, "crx.cyclic2");
		updateXmlNodeFromTextField(ui.m_ins, permutation, "mut.ins");
		updateXmlNodeFromTextField(ui.m_inv, permutation, "mut.inv");
		updateXmlNodeFromTextField(ui.m_toggle, permutation, "mut.toggle");

		// write back
		ofstream fout(PARAMS);
		if (fout.is_open())
		{
			fout << xConfig.createXMLString(true);
			fout.close();
		}
		else throw runtime_error("Cannot open parameters file.");
	}

	void CuttingStockApp::setTextFieldFromNode(QLineEdit* textField, const XMLNode& parent, const char* attribute)
	{
		textField->setText(parent.getChildNodeWithAttribute("Entry", "key", attribute).getText());
	}

	void CuttingStockApp::updateXmlNodeFromTextField(QLineEdit* textField, const XMLNode& parent, const char* attribute)
	{
		auto text = textField->text().toStdString();
		parent.getChildNodeWithAttribute("Entry", "key", attribute).updateText(text.c_str());
	}
}