#include "CuttingStockEvalOp.h"
#include "cuttingAlgorithm.h"
#include <boost/algorithm/string/find.hpp>

using namespace boost;
using cutting_stock::blf;

namespace cutting_stock
{
	void CuttingStockEvalOp::registerParameters(StateP state)
	{
		state->getRegistry()->registerEntry("tsp.infile", voidP(new string), ECF::STRING);
	}

	bool CuttingStockEvalOp::initialize(StateP state)
	{
		if (!state->getRegistry()->isModified("tsp.infile")) {
			state->getLogger()->log(1, "Error: no input file defined for TSP! (parameter 'tsp.infile'");
			return false;
		}

		auto sptr = state->getRegistry()->getEntry("tsp.infile"); // get parameter value
		const auto filePath = *static_cast<string*>(sptr.get()); // convert from voidP to user defined type

		ifstream iFile(filePath.c_str());
		string line;
		if (!iFile.is_open()) {
			state->getLogger()->log(1, "Error: can't open input file " + filePath);
			return false;
		}

		totalArea = 0;
		elements.clear();

		// read size (number of elements)
		while (getline(iFile, line) && line.find("size", 0) == string::npos);
		stringstream ss(line.substr(line.find(":") + 1));
		ss >> dimension;
		// set dimension for permutation genotype
		state->getRegistry()->modifyEntry("Permutation.size", voidP(new uint(dimension)));
		// reinitialize population with updated size
		state->getPopulation()->initialize(state);

		// parse data
		while (getline(iFile, line) && line.find("width", 0) == string::npos);
		auto r = find_nth(line, ":", 1);
		stringstream ssWidth(line.substr(distance(line.begin(), r.begin()) + 1));
		ssWidth >> width;

		while (getline(iFile, line) && line.find("quantity", 0) == string::npos);
		vector<int> xCoordinates, yCoordinates;
		int number, quantity;
		char c;
		while (iFile >> number) {
			iFile >> quantity;
			xCoordinates.clear();
			yCoordinates.clear();
			// repeat 2 times
			for (auto i = 0; i < 2; ++i) {
				iFile >> c;
				// read x coordinate
				if (c == 'x')
					parseCoordinates(iFile, xCoordinates);
				// read y coordinate
				else if (c == 'y')
					parseCoordinates(iFile, yCoordinates);
			}

			// add polygon to list quantity times
			for (auto i = 0; i < quantity; ++i) {
				Polygon polygon;
				polygon.number = number;

				// XCoordinates and YCoordinates must be of same length
				const auto len = xCoordinates.size();
				for (auto j = 0; j < len; ++j) {
					polygon.addPoint(xCoordinates[j], yCoordinates[j]);
				}

				polygon.setWidth();
				polygon.setHeight();
				polygon.setArea();
				totalArea += polygon.area;

				elements.push_back(polygon);
			}
		}

		iFile.close();
		return true;
	}

	// blf
	FitnessP CuttingStockEvalOp::evaluate(IndividualP individual)
	{
		FitnessP fitness(new FitnessMax);
		const auto perm = boost::static_pointer_cast<Permutation::Permutation> (individual->getGenotype());

		const auto fitnessV = blf(shared_from_this(), perm->variables);
		fitness->setValue(fitnessV);
		return fitness;
	}

	// provjerava moguca preklapanja s vec smjestenim elementima i prelazenje granica materijala
	bool CuttingStockEvalOp::canBePlaced(Polygon p, const Polygon::Point &point, const vector<Polygon> &placedElements) const
	{
		// ako stavimo p na point, postoji li preklapanje s bilo kojim smjestenim poligonom
		p.adjust(point.x, point.y);
		// x + w <= wmax
		if (p.coordinates[0].x + p.width > width) return false;
		const auto& size = placedElements.size();
		for (auto i = 0; i < size; ++i) {
			const auto& placedP = placedElements[i];
			// provjeri dolazi li do preklapanja
			if (rectanglesOverlap(p, placedP)) return false;
		}

		return true;
	}

	void CuttingStockEvalOp::parseCoordinates(ifstream &iFile, vector<int> &coordinates) {
		int c;
		while (iFile.peek() != '\n' && iFile >> c) {
			coordinates.push_back(c);
		}
	}

	bool CuttingStockEvalOp::rectanglesOverlap(const Polygon &rectA, const Polygon &rectB) {
		const auto& pA1 = rectA.coordinates[0];
		const auto& pB1 = rectB.coordinates[0];
		const auto& pA3 = rectA.coordinates[2];
		const auto& pB3 = rectB.coordinates[2];
		if (pA3.x < pB1.x || pB3.x < pA1.x || pA3.y < pB1.y || pB3.y < pA1.y) return false;
		return true;
	}
}
